<?php
/**
 * This source file is proprietary and part of Rebilly.
 *
 * (c) Rebilly SRL
 *     Rebilly Ltd.
 *     Rebilly Inc.
 *
 * @see https://www.rebilly.com
 */

declare(strict_types=1);

namespace Rebilly\Wp\Member\Action;

use Rebilly\Entities\Customer;
use Rebilly\Entities\PaymentCard;
use Rebilly\Entities\PaymentCardToken;
use Rebilly\Http\Exception\UnprocessableEntityException;
use Rebilly\Wp\Member\Entity\User;
use Rebilly\Wp\Member\Exception\AjaxFailException;
use Rebilly\Wp\Member\Plugin;

// Exit if accessed directly.
if (!defined('ABSPATH')) {
    exit();
}

/**
 * Class CardAdd
 * @package Rebilly\Wp\Member\Action
 */
class CardAdd extends AjaxAction
{
    public function __construct()
    {
        parent::__construct('rebilly_card_add');
    }

    /**
     * @param array $data
     *
     * @throws AjaxFailException
     * @throws \Rebilly\Wp\Member\Exception\AjaxSuccessException
     * @return array
     */
    public function process(array $data): array
    {
        $user = User::find();
        if (!$user) {
            $this->fail('You must be authorized to add any payment card');
        }

        if ($this->addCard($user, $data)) {
            $this->success('Your payment card is added');
        }

        return [
            'success' => false,
            'message' => __('Unexpected error, try again', REBILLY_WP_LANG_DOMAIN),
        ];
    }

    /**
     * @param User $user
     * @param array $rawData
     *
     * @throws AjaxFailException
     * @return bool
     */
    private function addCard(User $user, array $rawData): bool
    {
        $framePayToken = $rawData['rebilly-token'];
        $token = $this->getToken($framePayToken);

        $this->maybeCreateCustomer($user, $token);

        return true;
    }

    /**
     * @param User|null $user
     * @param PaymentCardToken $token
     *
     * @throws AjaxFailException
     * @return Customer
     */
    private function maybeCreateCustomer(User $user, PaymentCardToken $token): Customer
    {
        $api = Plugin::api();

        /** @var bool|Customer $customer */
        $customer = false;
        $customerExisted = true;
        if ($customerId = $user->getCustomerId()) {
            if (!($customer = $api->customer()->get((string) $customerId))) {
                // Nothing we should do. Just create new customer, but clear this field before
                // so we will not try this again
                $user->setCustomerId(false);
            }
        }
        if (!$customer) {
            try {
                $customers = $api->customer()->findByEmail($user->user_email);
                if ($customers->count()) {
                    $customer = $customers->offsetGet(0);
                }

                if (!$customer) {
                    // Create new one if he is not found by email address
                    $customer = $api->customer()->createFromToken($token->getId());
                    $customerExisted = false;
                }
            } catch (UnprocessableEntityException $e) {
                // Can only be validation error, so just print it out
                $this->fail('Please fill in the form', ['form' => $e->getErrors()]);
            }
        }

        if ($customerExisted) {
            $this->createPaymentCard($customer, $token);
        }

        // We created/changed customerId and need to save it
        if ($customerId !== $customer->getId()) {
            $user->setCustomerId($customer->getId());
        }

        return $customer;
    }

    /**
     * @param Customer $customer
     * @param PaymentCardToken $token
     *
     * @throws AjaxFailException
     * @return bool
     */
    private function createPaymentCard(Customer $customer, PaymentCardToken $token): bool
    {
        $api = Plugin::api();

        try {
            if ($customer->getDefaultPaymentInstrument() === null) {
                $customer->setPaymentToken($token->getId());
                $api->customer()
                    ->update($customer->getId(), $customer);
            } else {
                $api->paymentCard()
                    ->create(
                        (new PaymentCard())
                            ->setCustomerId($customer->getId())
                            ->setToken($token->getId())
                    );
            }
        } catch (UnprocessableEntityException $e) {
            $this->fail($e->getMessage(), $e->getErrors());
        }

        return true;
    }

    /**
     * @param string $token
     *
     * @throws AjaxFailException
     * @return PaymentCardToken
     */
    private function getToken(string $token): PaymentCardToken
    {
        $entity = Plugin::api()
            ->paymentCard()
            ->getToken($token);

        if (!$entity || $entity->getIsUsed()) {
            // Second time we try to create PaymentCard from token
            $this->fail('Payment card data expired, please try again');
        }

        return $entity;
    }
}
