<?php
/**
 * This source file is proprietary and part of Rebilly.
 *
 * (c) Rebilly SRL
 *     Rebilly Ltd.
 *     Rebilly Inc.
 *
 * @see https://www.rebilly.com
 */

declare(strict_types=1);

namespace Rebilly\Wp\Member\Action;

use Rebilly\Entities\Plan;
use Rebilly\Entities\Subscriptions\PlanPricing;
use Rebilly\Entities\Subscriptions\Pricing\FixedFee;
use Rebilly\Entities\Subscriptions\Pricing\FlatRate;
use Rebilly\Entities\Subscriptions\Pricing\Stairstep;
use Rebilly\Entities\Subscriptions\Pricing\Tiered;
use Rebilly\Entities\Subscriptions\Pricing\Volume;
use Rebilly\Wp\Member\Entity\User;
use Rebilly\Wp\Member\Plugin;

// Exit if accessed directly.
if (!defined('ABSPATH')) {
    exit();
}

class SubscriptionsPaginate extends AjaxAction
{
    public function __construct()
    {
        parent::__construct('rebilly_subscriptions_paginate', false);
    }

    /**
     * @param array $data
     * @return array
     */
    public function process(array $data): array
    {
        $user = User::find();

        $perPage = (int) apply_filters('rebilly_subscriptions_per_page', 10);
        $page = max((int) ($data['page'] ?? 1), 1);

        $subscriptions = [];

        if ($user->getCustomerId()) {
            $subscriptionsRaw = Plugin::api()
                ->subscription()
                ->findByCustomer($user->getCustomerId(), $page, $perPage);
            $planIds = [];
            foreach ($subscriptionsRaw as $subscription) {
                $plans = [];
                foreach ($subscription->getItems() as $item) {
                    $planIds[] = $item->getPlanId();
                    $plans[$item->getPlanId()] = [
                        'id' => $item->getPlanId(),
                        'quantity' => $item->getQuantity(),
                    ];
                }
                $subscriptions[$subscription->getId()] = [
                    'id' => $subscription->getId(),
                    'status' => $subscription->getStatus(),
                    'createdTime' => strtotime($subscription->getCreatedTime() ?? ''),
                    'startTime' => strtotime($subscription->getStartTime() ?? ''),
                    'renewalTime' => strtotime($subscription->getRenewalTime() ?? ''),
                    'canceledTime' => strtotime($subscription->getCanceledTime() ?? ''),
                    'plans' => $plans,
                ];
            }
            $planIds = array_unique($planIds);
            $plansRaw = Plugin::api()
                ->plan()
                ->all($planIds);
            /** @var Plan[] $plans */
            $plans = [];
            foreach ($plansRaw as $plan) {
                $plans[$plan->getId()] = $plan;
            }
            // TODO trial and setup support
            foreach ($subscriptions as $subscriptionId => $subscription) {
                foreach ($subscription['plans'] as $planId => $planRaw) {
                    $plan = &$plans[$planId];
                    $recurringInterval = $plan->getRecurringInterval();
                    $subscriptions[$subscriptionId]['plans'][$planId] = array_merge($planRaw, [
                        'name' => $plan->getName(),
                        'recurring' => $recurringInterval ? [
                            'unit' => $plan->getRecurringInterval()->getUnit(),
                            'length' => $plan->getRecurringInterval()->getLength(),
                        ] : false,
                        'price' => $this->extractPrice($plan->getPricing(), $planRaw['quantity']),
                        'currency' => $plan->getCurrency(),
                        'currencySign' => $plan->getCurrencySign(),
                    ]);
                }
            }
            $pageCount = ceil($subscriptionsRaw->getTotalItems() / $perPage);
        }

        $actions = [
            'cancel' => Action::get('rebilly_subscription_cancel'),
        ];

        return [
            'success' => true,
            'items' => renderIsolate(REBILLY_WP_DIR . 'template/front/ajax/subscription-list.php', compact('subscriptions', 'actions'), true),
            'page' => $page,
            'pageCount' => $pageCount ?? 0,
        ];
    }

    /**
     * @param PlanPricing|FlatRate|FixedFee|Stairstep|Tiered|Volume $pricing
     * @param int $quantity
     *
     * @return float
     */
    private function extractPrice(PlanPricing $pricing, int $quantity): float
    {
        switch ($pricing->getFormula()) {
            case PlanPricing::FLAT_RATE:
                return $pricing->getPrice() * $quantity;
            case PlanPricing::STAIRSTEP:
                foreach ($pricing->getBrackets() as $bracket) {
                    if ($quantity <= $bracket->getMaxQuantity()) {
                        return $bracket->getPrice();
                    }
                }

                return 0;
            case PlanPricing::TIERED:
                $sum = 0;
                foreach ($pricing->getBrackets() as $bracket) {
                    $sum += $bracket->getPrice() * min($bracket->getMaxQuantity(), $quantity);
                    $quantity -= min($bracket->getMaxQuantity(), $quantity);
                    if ($quantity <= 0) {
                        break;
                    }
                }

                return $sum;
            case PlanPricing::VOLUME:
                $price = 0;
                foreach ($pricing->getBrackets() as $bracket) {
                    if ($quantity <= $bracket->getMaxQuantity()) {
                        $price = $bracket->getPrice();

                        break;
                    }
                }

                return $price * $quantity;
            case PlanPricing::FIXED_FEE:
            default:
                return $pricing->getPrice();
        }
    }
}
