<?php
/**
 * This source file is proprietary and part of Rebilly.
 *
 * (c) Rebilly SRL
 *     Rebilly Ltd.
 *     Rebilly Inc.
 *
 * @see https://www.rebilly.com
 */

declare(strict_types=1);

namespace Rebilly\Wp\Member\Api;

use Rebilly\Client as Api;
use Rebilly\Middleware\CompositeMiddleware;
use Rebilly\Wp\Member\Api\Middleware\HandleUnauthorized;
use Rebilly\Wp\Member\Api\Middleware\UserAgent;

// Exit if accessed directly.
if (!defined('ABSPATH')) {
    exit();
}

/**
 * Class Client
 * @package Rebilly\Wp\Member\Api
 */
class Client
{
    public const UNAUTHORIZED_EXPIRE = 0;

    /** @var string */
    private $secretKey;

    /** @var bool */
    private $isUnauthorized;

    /** @var bool */
    private $isSandbox;

    /** @var Api */
    private $api;

    /** @var string */
    private $websiteId;

    /** @var Service[]|array */
    private $services;

    /**
     * Client constructor.
     *
     * @param string $secretKey
     * @param string $websiteId
     * @param bool $sandbox
     */
    public function __construct(string $secretKey, string $websiteId, bool $sandbox)
    {
        $this->secretKey = $secretKey;
        $this->websiteId = $websiteId;
        $this->isSandbox = $sandbox;

        $this->api = new Api([
            'apiKey' => $secretKey,
            'baseUrl' => $sandbox ? Api::SANDBOX_HOST : Api::BASE_HOST,
            'middleware' => new CompositeMiddleware(
                new UserAgent(),
                new HandleUnauthorized($this)
            ),
        ]);

        $this->services = [
            'customer' => new CustomerService($this),
            'customField' => new CustomFieldService($this),
            'invoice' => new InvoiceService($this),
            'paymentCard' => new PaymentCardService($this),
            'plan' => new PlanService($this),
            'product' => new ProductService($this),
            'rulesEngine' => new RulesEngineService($this),
            'subscription' => new SubscriptionService($this),
            'transaction' => new TransactionService($this),
            'webhook' => new WebhookService($this),
            'website' => new WebsiteService($this),
        ];

        $this->isUnauthorized = $this->checkIsUnauthorized();
    }

    /**
     * @return CustomerService
     */
    public function customer(): CustomerService
    {
        return $this->services['customer'];
    }

    /**
     * @return CustomFieldService
     */
    public function customField(): CustomFieldService
    {
        return $this->services['customField'];
    }

    /**
     * @return InvoiceService
     */
    public function invoice(): InvoiceService
    {
        return $this->services['invoice'];
    }

    /**
     * @return PaymentCardService
     */
    public function paymentCard(): PaymentCardService
    {
        return $this->services['paymentCard'];
    }

    /**
     * @return PlanService
     */
    public function plan(): PlanService
    {
        return $this->services['plan'];
    }

    /**
     * @return ProductService
     */
    public function product(): ProductService
    {
        return $this->services['product'];
    }

    /**
     * @return RulesEngineService
     */
    public function rulesEngine(): RulesEngineService
    {
        return $this->services['rulesEngine'];
    }

    /**
     * @return SubscriptionService
     */
    public function subscription(): SubscriptionService
    {
        return $this->services['subscription'];
    }

    /**
     * @return TransactionService
     */
    public function transaction(): TransactionService
    {
        return $this->services['transaction'];
    }

    /**
     * @return WebhookService
     */
    public function webhook(): WebhookService
    {
        return $this->services['webhook'];
    }

    /**
     * @return WebsiteService
     */
    public function website(): WebsiteService
    {
        return $this->services['website'];
    }

    /**
     * @return Api
     */
    public function api(): Api
    {
        return $this->api;
    }

    public function isConnected(): bool
    {
        return $this->secretKey !== '' && !$this->isUnauthorized;
    }

    /**
     * @return string
     */
    public function getWebsiteId(): string
    {
        return $this->websiteId;
    }

    public function setUnauthorized($isUnauthorized = true): void
    {
        $this->isUnauthorized = $isUnauthorized;
        $transientKey = 'rebilly__unauthorized';
        if ($isUnauthorized) {
            set_transient(
                $transientKey,
                sha1("{$this->websiteId}-{$this->secretKey}" . ($this->isSandbox ? '-sandbox' : '')),
                self::UNAUTHORIZED_EXPIRE
            );
        } else {
            delete_transient($transientKey);
        }
    }

    private function checkIsUnauthorized(): bool
    {
        $transientKey = 'rebilly__unauthorized';
        $value = get_transient($transientKey);
        if ($value !== false) {
            $result = $value === sha1("{$this->websiteId}-{$this->secretKey}" . ($this->isSandbox ? '-sandbox' : ''));
            if ($result === false) {
                delete_transient($transientKey);
            }

            return $result;
        }

        return false;
    }
}
