<?php
/**
 * This source file is proprietary and part of Rebilly.
 *
 * (c) Rebilly SRL
 *     Rebilly Ltd.
 *     Rebilly Inc.
 *
 * @see https://www.rebilly.com
 */

declare(strict_types=1);

namespace Rebilly\Wp\Member\Api;

use Rebilly\Entities\CustomField;
use Rebilly\Http\Exception\ClientException;
use Rebilly\Http\Exception\NotFoundException;
use Rebilly\Http\Exception\UnprocessableEntityException;
use Rebilly\Rest\Collection;

// Exit if accessed directly.
if (!defined('ABSPATH')) {
    exit();
}

/**
 * Class CustomFieldsService
 * @package Rebilly\Wp\Member\Api
 */
class CustomFieldService extends Service
{
    /**
     * @param string $resource
     * @param string $name
     *
     * @return CustomField
     * @throws NotFoundException
     */
    public function get(string $resource, string $name): CustomField
    {
        return $this->api()
            ->customFields()
            ->load($resource, $name);
    }

    /**
     * @param string $resource
     * @param int $cache
     * @param bool $force
     *
     * @return CustomField[]|Collection
     */
    public function all(string $resource, int $cache = 5 * 60, bool $force = false): Collection
    {
        return $this->cached(
            "$resource-custom-fields",
            function () use ($resource) {
                try {
                    return $this->api()
                        ->customFields()
                        ->search($resource);
                } catch (ClientException $ex) {
                }

                return false;
            },
            $cache,
            $force
        ) ?: new Collection(new CustomField());
    }

    /**
     * @param string $resource
     * @param int $cache
     * @param bool $force
     *
     * @return string[]|array
     */
    public function allByName(string $resource, int $cache = 5 * 60, bool $force = false): array
    {
        return $this->cached(
            "$resource-custom-fields",
            function () use ($resource, $cache, $force) {
                $list = [];
                $rawFields = $this->all($resource, $cache, $force);
                foreach ($rawFields as $field) {
                    $fieldName = $field->getName();
                    $list[$fieldName] = $field->jsonSerialize();
                }

                return count($list) > 0 ? $list : false;
            },
            $cache,
            $force
        ) ?: [];
    }

    /**
     * @param string $resource
     * @param string $name
     * @param CustomField|array $data
     *
     * @return CustomField
     * @throws UnprocessableEntityException
     */
    public function create(string $resource, string $name, $data): CustomField
    {
        return $this->api()
            ->customFields()
            ->create($resource, $name, $data);
    }

    /**
     * @param string $websiteId
     *
     * @return array
     * @throws UnprocessableEntityException
     */
    public function registerRulesEngine(string $websiteId): array
    {
        $fields = [];
        $toCreate = [
            'customers' => self::customerFields(),
            'subscriptions' => self::subscriptionFields(),
        ];
        foreach ($toCreate as $resource => $fieldset) {
            foreach ($fieldset as $role => $field) {
                $name = str_replace('{websiteId}', $websiteId, $field['name']);
                $name = preg_replace('/[^A-Za-z0-9_-]/', '_', $name);

                $customField = $this->api()
                    ->customFields()
                    ->update($resource, $name, $field['field']);

                $fields[$role] = $customField->getName();
            }
        }

        return $fields;
    }

    public static function customerFields(): array
    {
        return [
            'username' => [
                'name' => 'wp-{websiteId}-username',
                'field' => (new CustomField())
                    ->setType(CustomField::TYPE_STRING)
                    ->setDescription('WordPress username'),
            ],
        ];
    }

    public static function subscriptionFields(): array
    {
        return [
            'is-wp' => [
                'name' => 'wp-source',
                'field' => (new CustomField())
                    ->setType(CustomField::TYPE_BOOLEAN)
                    ->setDescription('Is created from WordPress'),
            ],
            'email-confirmation' => [
                'name' => 'wp-email-confirmation',
                'field' => (new CustomField())
                    ->setType(CustomField::TYPE_BOOLEAN)
                    ->setDescription('Is WordPress confirmation email required'),
            ],
            'email-confirmation-url' => [
                'name' => 'wp-email-confirmation-url',
                'field' => (new CustomField())
                    ->setType(CustomField::TYPE_STRING)
                    ->setDescription('WordPress confirmation email URL'),
            ],
        ];
    }
}
