<?php
/**
 * This source file is proprietary and part of Rebilly.
 *
 * (c) Rebilly SRL
 *     Rebilly Ltd.
 *     Rebilly Inc.
 *
 * @see https://www.rebilly.com
 */

declare(strict_types=1);

namespace Rebilly\Wp\Member\Api;

use Exception;
use Rebilly\Entities\PaymentCard;
use Rebilly\Entities\PaymentCardToken;
use Rebilly\Http\Exception\UnprocessableEntityException;
use Rebilly\Rest\Collection;

// Exit if accessed directly.
if (!defined('ABSPATH')) {
    exit();
}

/**
 * Class PaymentCardService
 * @package Rebilly\Wp\Member\Api
 */
class PaymentCardService extends Service
{
    /**
     * @param string $id
     *
     * @return PaymentCard|null
     */
    public function get(string $id): ?PaymentCard
    {
        if (!blank($id)) {
            try {
                return $this->api()
                    ->paymentCards()
                    ->load($id);
            } catch (Exception $ex) {
            }
        }

        return null;
    }

    /**
     * @param string $tokenId
     *
     * @return PaymentCardToken|null
     */
    public function getToken(string $tokenId): ?PaymentCardToken
    {
        if (!blank($tokenId)) {
            try {
                return $this->api()
                    ->paymentCardTokens()
                    ->load($tokenId);
            } catch (Exception $ex) {
            }
        }

        return null;
    }

    /**
     * @param array $params
     *
     * @return PaymentCard[]|Collection
     */
    public function find(array $params): Collection
    {
        try {
            return $this->api()
                ->paymentCards()
                ->search($params);
        } catch (Exception $ex) {
        }

        return new Collection(new PaymentCard());
    }

    /**
     * @param string $customerId
     *
     * @return PaymentCard[]|Collection
     */
    public function findByCustomer(string $customerId): Collection
    {
        return $this->find([
            'filter' => "customerId:$customerId;status:!deactivated",
        ]);
    }

    /**
     * @param PaymentCard|array $data
     *
     * @throws UnprocessableEntityException
     * @return PaymentCard|null
     */
    public function create($data): ?PaymentCard
    {
        try {
            return $this->api()
                ->paymentCards()
                ->create($data);
        } catch (Exception $ex) {
            if ($ex instanceof UnprocessableEntityException) {
                throw $ex;
            }
        }

        return null;
    }

    /**
     * @param string $customerId
     * @param string $tokenId
     *
     * @throws UnprocessableEntityException
     * @return PaymentCard|null
     */
    public function createFromToken(string $customerId, string $tokenId): ?PaymentCard
    {
        return $this->create(
            (new PaymentCard())
                ->setCustomerId($customerId)
                ->setToken($tokenId)
        );
    }

    /**
     * @param string $id
     * @param PaymentCard|array $data
     *
     * @return PaymentCard|null
     */
    public function update(string $id, $data): ?PaymentCard
    {
        try {
            return $this->api()
                ->paymentCards()
                ->update($id, $data);
        } catch (Exception $ex) {
        }

        return null;
    }

    /**
     * @param string $id
     *
     * @return PaymentCard|null
     */
    public function deactivate(string $id): ?PaymentCard
    {
        try {
            return $this->api()
                ->paymentCards()
                ->deactivate($id);
        } catch (Exception $ex) {
        }

        return null;
    }
}
