<?php
/**
 * This source file is proprietary and part of Rebilly.
 *
 * (c) Rebilly SRL
 *     Rebilly Ltd.
 *     Rebilly Inc.
 *
 * @see https://www.rebilly.com
 */

namespace Rebilly\Wp\Member\Entity;

use Rakit\Validation\Rules\In;
use Rakit\Validation\Validation;
use Rakit\Validation\Validator;
use Rebilly\Entities\CustomField;
use Rebilly\Wp\Member\Reformer\Reformer;
use Rebilly\Wp\Member\Reformer\Select;
use WP_Post;

// Exit if accessed directly.
if (!defined('ABSPATH')) {
    exit();
}

/**
 * Class Form
 * @package Rebilly\Wp\Member\Entity
 *
 * @mixin WP_Post
 */
class Form
{
    use PostProxy, HasMeta, HasDomTree;

    public const ROLE_LOGIN = 'login';

    public const ROLE_SUBSCRIPTION = 'subscription';

    public const ROLE_ADD_CARD = 'add-card';

    public const FIELD_MAP = [
        'firstName' => ['billingAddress.firstName', 'deliveryAddress.firstName'],
        'lastName' => ['billingAddress.lastName', 'deliveryAddress.lastName'],
        'address' => 'billingAddress.address',
        'address2' => 'billingAddress.address2',
        'city' => 'billingAddress.city',
        'region' => 'billingAddress.region',
        'country' => 'billingAddress.country',
        'postalCode' => 'billingAddress.postalCode',
        'delivery_address' => 'deliveryAddress.address',
        'delivery_address2' => 'deliveryAddress.address2',
        'delivery_city' => 'deliveryAddress.city',
        'delivery_region' => 'deliveryAddress.region',
        'delivery_country' => 'deliveryAddress.county',
        'delivery_postalCode' => 'deliveryAddress.postalCode',
        'emails' => ['billingAddress.emails', 'deliveryAddress.emails'],
        'phoneNumbers' => ['billingAddress.phoneNumbers', 'deliveryAddress.phoneNumbers'],
    ];

    protected static $postType = 'rebilly_m_form';

    public static $meta = [
        'role' => '_rebilly_role',
        'plan' => '_rebilly_m_plan_id',
        'allow-guest' => '_rebilly_m_allow_guest',
        'enable-signup' => '_rebilly_m_enable_signup',
        'confirmation-email' => '_rebilly_m_confirmation_email',
        'page-success' => '_rebilly_m_page_success',
        'page-pending' => '_rebilly_m_page_pending',
        'addons' => '_rebilly_addons',
        'bump' => '_rebilly_bump',
        'fields' => '_rebilly_m_fields',
    ];

    public static function register(): void
    {
        add_action('init', [static::class, 'init']);
    }

    public static function init(): void
    {
        register_post_type(
            static::$postType,
            [
                'label' => __('Forms', REBILLY_WP_LANG_DOMAIN),
                'labels' => [
                    'name' => __('Forms', REBILLY_WP_LANG_DOMAIN),
                    'singular_name' => __('Form', REBILLY_WP_LANG_DOMAIN),
                    'menu_name' => __('Forms', REBILLY_WP_LANG_DOMAIN),
                    'all_items' => __('All Forms', REBILLY_WP_LANG_DOMAIN),
                    'add_new' => __('Add Form', REBILLY_WP_LANG_DOMAIN),
                    'add_new_item' => __('Add Form', REBILLY_WP_LANG_DOMAIN),
                    'edit_item' => __('Edit Form', REBILLY_WP_LANG_DOMAIN),
                    'new_item' => __('New Forms', REBILLY_WP_LANG_DOMAIN),
                    'view_item' => __('View Form', REBILLY_WP_LANG_DOMAIN),
                    'view_items' => __('View Forms', REBILLY_WP_LANG_DOMAIN),
                    'search_items' => __('Search Form', REBILLY_WP_LANG_DOMAIN),
                    'not_found' => __('No Forms found', REBILLY_WP_LANG_DOMAIN),
                    'not_found_in_trash' => __('No Forms found in Trash', REBILLY_WP_LANG_DOMAIN),
                ],
                'description' => __('Rebilly Form', REBILLY_WP_LANG_DOMAIN),
                'public' => true,
                'publicly_queryable' => false,
                'show_ui' => true,
                'delete_with_user' => false,
                'show_in_rest' => true,
                'has_archive' => false,
                'show_in_menu' => false,
                'show_in_nav_menus' => false,
                'exclude_from_search' => true,
                'capability_type' => 'post',
                'map_meta_cap' => true,
                'hierarchical' => false,
                'rewrite' => ['slug' => static::$postType, 'with_front' => true],
                'query_var' => true,
                'supports' => ['title'],
            ]
        );
    }

    public static function roles(): array
    {
        return [
            self::ROLE_SUBSCRIPTION => 'Subscription',
            self::ROLE_LOGIN => 'Login',
            self::ROLE_ADD_CARD => 'Add Payment Card',
        ];
    }

    /**
     * @param array $args
     *
     * @return WP_Post[]|null|array
     */
    public static function all($args = []): ?array
    {
        return get_posts(wp_parse_args($args, [
            'post_type' => static::$postType,
            'posts_per_page' => -1,
        ]));
    }

    public function metaDefaults(): array
    {
        return [
            'role' => self::ROLE_SUBSCRIPTION,
            'plan' => [],
            'allow-guest' => true,
            'enable-signup' => false,
            'confirmation-email' => true,
            'page-success-type' => true,
            'page-pending-type' => true,
        ];
    }

    public function getPage($name)
    {
        $pageConfig = $this->metaGet("page-{$name}", true);
        if ($pageConfig['type'] === 'post') {
            return $pageConfig['id'] ? get_post($pageConfig['id']) : null;
        }

        return $pageConfig['url'];
    }

    public function getPageUrl($name)
    {
        $page = $this->getPage($name);

        return $page instanceof WP_Post ? get_permalink($page) : $page;
    }

    public static function getCustomFields(): array
    {
        $subscription = Subscription::getCustomFields();
        $customer = User::getCustomFields();

        return compact('subscription', 'customer');
    }

    public static function getCustomFieldsOptions(): array
    {
        $customFields = [];
        $raw = self::getCustomFields();
        foreach ($raw as $group => $customFieldsGroup) {
            $groupName = __(mb_convert_case($group, MB_CASE_TITLE), REBILLY_WP_LANG_DOMAIN);
            foreach ($customFieldsGroup as $key => $field) {
                $customFields["$group-$key"] = "{$field['description']} ($groupName)";
            }
        }

        return $customFields;
    }

    /**
     * @param bool $isBump
     * @return string[]|array
     */
    public function getPlanIds($isBump = false): array
    {
        return $this->metaGet($isBump ? 'bump' : 'plan', true, []);
    }

    /**
     * @return bool
     */
    public function isGuestAllowed(): bool
    {
        return (bool) $this->metaGet('allow-guest', true, true);
    }

    /**
     * @return bool
     */
    public function isSignupEnabled(): bool
    {
        return (bool) $this->metaGet('enable-signup', true);
    }

    /**
     * @return bool
     */
    public function isRebillyConfirmationStrategy(): bool
    {
        return (bool) $this->metaGet('confirmation-email', true);
    }

    public function isAuthenticationRequired(): bool
    {
        switch ($this->getRole()) {
            case self::ROLE_SUBSCRIPTION:
                return !$this->isGuestAllowed();
            case self::ROLE_ADD_CARD:
                return true;
            case self::ROLE_LOGIN:
                return false;
        }

        return true;
    }

    /**
     * @return string
     */
    public function getRole(): string
    {
        return (string) $this->metaGet('role', true, self::ROLE_SUBSCRIPTION);
    }

    /**
     * @param string $addon
     * @return string[]|array
     */
    public function getAddonPlanIds(string $addon = ''): array
    {
        $addons = $this->metaGet('addons', true, []);
        $result = [];
        if (!blank($addon)) {
            foreach ((array) ($addons[$addon] ?? []) as $planId) {
                $result[$planId] = 1;
            }
        } else {
            foreach ($addons as $addonRow) {
                foreach ((array) $addonRow as $planId) {
                    $result[$planId] = 1;
                }
            }
        }

        return $result;
    }

    /**
     * @param array|null $components
     * @return array
     */
    public function getPlanQuantityFields(?array $components = null): array
    {
        $fields = $this->getFields($components);
        $quantities = [];
        foreach ($fields as $name => $field) {
            if (!blank($field['settings']['plan'] ?? '')) {
                $quantities[$field['settings']['plan']] = $name;
            }
        }

        return $quantities;
    }

    /**
     * @param array|null $components
     * @return array
     */
    public function getAddonFields(?array $components = null): array
    {
        $fields = $this->getFields($components);
        $addons = [];
        foreach ($fields as $name => $field) {
            if (!blank($field['settings']['addon'] ?? '')) {
                $addons[$name] = $field;
            }
        }

        return $addons;
    }

    /**
     * @param array|null $components
     *
     * @return array
     */
    public function getBumpFields(?array $components = null): array
    {
        $fields = $this->getFields($components);
        $bumps = [];
        foreach ($fields as $name => $field) {
            if ($field['settings']['bump'] ?? false) {
                $bumps[$name] = $field;
            }
        }

        return $bumps;
    }

    /**
     * @param array|null $components
     * @param string|null $condition
     *
     * @return array
     */
    public function getFields(?array $components = null, ?string $condition = null): array
    {
        if ($components === null) {
            return $this->getFields($this->getComponents());
        }
        if (isset($components['type'])) {
            $components = [$components];
        }
        /** @var array|bool $label */
        $label = false;
        $fields = [];
        $field_groups = [[]];
        foreach ($components as $component) {
            if (in_array($component['type'] ?? '', ['input', 'number', 'textarea', 'select', 'checkbox', 'radio'], true) &&
                !blank($component['attributes']['name'] ?? '')) {
                unset($component['children']);
                if ($label !== false && !blank($text = Reformer::renderText($label))) {
                    $component['label'] = $text;
                }
                $component['settings'] = $component['settings'] ?? [];
                $component['settings']['condition'] = $component['settings']['condition'] ?? $condition;
                $fields[$component['attributes']['name']] = $component;
                if ($component['type'] === 'select') {
                    $select = new Select($component);
                    $fields[$component['attributes']['name']]['values'] = $select->getOptions();
                }
            } else {
                $label = $component['type'] === 'label' ? $component : false;
                $field_groups[] = $this->getFields($component['children'] ?? [], $component['settings']['condition'] ?? $condition);
            }
        }

        return array_merge($fields, ...$field_groups);
    }

    public function getComponents(): array
    {
        $fields = $this->metaGet('fields', true);
        $fields = json_decode($fields, true);

        return $this->buildComponents($fields['components'] ?? []);
    }

    public function getStyles(): array
    {
        $fields = $this->metaGet('fields', true);
        $fields = json_decode($fields, true);

        return $fields['styles'] ?? [];
    }

    private function isConditionSatisfied(array $condition, array $data): bool
    {
        if (empty($condition)) {
            return true;
        }

        switch ($condition['op'] ?? 'equals') {
            case 'and':
                foreach ($condition['conditions'] ?? [] as $cond) {
                    if (!$this->isConditionSatisfied($cond, $data)) {
                        return false;
                    }
                }

                return true;
            case 'or':
                foreach ($condition['conditions'] ?? [] as $cond) {
                    if ($this->isConditionSatisfied($cond, $data)) {
                        return true;
                    }
                }

                return false;
            case 'not-equals':
                return ($data[$condition['name']] ?? '') != $condition['value'];
            case 'equals':
            default:
                return ($data[$condition['name']] ?? '') == $condition['value'];
        }
    }

    public function getRules(array $data): array
    {
        $user = User::find();
        $isAuthenticated = (bool) $user;

        $fields = $this->getFields();
        $ruleSet = [];
        foreach ($fields as $fieldName => $field) {
            $rules = [];
            $isRequired = $field['attributes']['required'] ?? false;
            $showFor = $field['settings']['show-for'] ?? false;
            $isConditionSatisfied = $this->isConditionSatisfied(json_decode($field['settings']['condition'] ?? '{}', true), $data);
            $isShown = !(($showFor === 'auth' && !$isAuthenticated) || ($showFor === 'guest' && $isAuthenticated));
            $isMandatory = in_array($field['settings']['rebilly'] ?? '', ['firstName', 'lastName', 'email'], true);
            if ($isMandatory || ($isShown && $isConditionSatisfied && $isRequired)) {
                $rules[] = 'required';
            }
            switch ($field['type']) {
                case 'input':
                    switch ($field['attributes']['type'] ?? '') {
                        case 'email':
                            $rules[] = 'email';

                            break;
                        case 'date':
                            $rules[] = 'date';

                            break;
                    }
                    if (array_key_exists('maxlength', $field['attributes']) && is_numeric($field['attributes']['maxlength'])) {
                        $rules[] = "min:{$field['attributes']['maxlength']}";
                    }

                    break;
                case 'number':
                    $rules[] = 'numeric';
                    $isQuantity = !blank($field['settings']['plan'] ?? ($field['attributes']['plan'] ?? ''));
                    if (array_key_exists('min', $field['attributes']) && is_numeric($field['attributes']['min'])) {
                        $rules[] = "min:{$field['attributes']['min']}";
                    } elseif ($isQuantity) {
                        $rules[] = 'min:1';
                    }
                    if (array_key_exists('max', $field['attributes']) && is_numeric($field['attributes']['max'])) {
                        $rules[] = "max:{$field['attributes']['max']}";
                    } elseif ($isQuantity) {
                        $rules[] = 'max:100';
                    }

                    break;
                case 'select':
                    $allowed = array_keys($field['values']);
                    $rule = (new In())->fillParameters($allowed);
                    if (array_key_exists('multiple', $field) && $field['multiple']) {
                        $rules[] = 'array';
                        $ruleSet[$fieldName . '.*'] = [$rule];
                    } else {
                        $rules[] = $rule;
                    }

                    break;
            }

            if (count($rules)) {
                $ruleSet[$fieldName] = $rules;
            }
        }

        return $ruleSet;
    }

    public function validate(array $data): Validation
    {
        $rules = $this->getRules($data);
        foreach ($data as $key => &$value) {
            if (is_numeric($value) &&
                array_key_exists($key, $rules) &&
                count(array_intersect(['numeric', 'integer'], array_filter($rules[$key], 'is_string') ?? [])) > 0) {
                $value += 0;
            }
        }
        unset($value);

        $validation = (new Validator())->make($data, $rules, [
            'required' => __(':attribute is required', REBILLY_WP_LANG_DOMAIN),
            'email' => __(':attribute must contain valid email', REBILLY_WP_LANG_DOMAIN),
            'numeric' => __(':attribute must be numeric', REBILLY_WP_LANG_DOMAIN),
            'min' => __(':attribute must be greater than or equal to :min', REBILLY_WP_LANG_DOMAIN),
            'max' => __(':attribute must be less than or equal to :max', REBILLY_WP_LANG_DOMAIN),
            'between' => __(':attribute must be between :min and :max', REBILLY_WP_LANG_DOMAIN),
            'in' => __(':attribute is invalid', REBILLY_WP_LANG_DOMAIN),
        ]);
        $validation->setAliases($this->getFieldAliases());
        $validation->validate();

        return $validation;
    }

    /**
     * Transforms form data to Rebilly subscription fields
     *
     * @param $data
     * @param null|string $email
     *
     * @return array
     */
    public function getRebillyFields(array $data, ?string &$email = null): array
    {
        $fields = [
            'billingAddress' => [],
            'deliveryAddress' => [],
            'emails' => [],
            'phoneNumbers' => [],
            'customFields' => [],
        ];
        $customFields = static::getCustomFields();
        foreach ($this->getFields() as $field) {
            $role = $field['settings']['rebilly'];
            if (blank($role)) {
                continue;
            }
            $fieldName = $field['attributes']['name'] ?? '';
            if (array_key_exists($fieldName, $data)) {
                $value = $data[$fieldName];
                $roleParts = explode('-', $role, 2);
                if ($role === 'email') {
                    $fields['emails'][] = [
                        'label' => empty($fields['emails']) ? 'main' : $field['label'],
                        'value' => $value,
                        'primary' => empty($fields['emails']),
                    ];
                    if (blank($email)) {
                        $email = $value;
                    }
                } elseif ($role === 'phone') {
                    $fields['phoneNumbers'][] = [
                        'label' => empty($fields['phoneNumbers']) ? 'main' : $field['label'],
                        'value' => $value,
                        'primary' => empty($fields['phoneNumbers']),
                    ];
                } elseif (count($roleParts) === 2) {
                    [$entity, $customFieldName] = $roleParts;
                    if (!isset($customFields[$entity][$customFieldName])) {
                        continue;
                    }
                    $value = $this->castCustomField($value, $customFields[$entity][$customFieldName]);
                    $fields['customFields'][$entity] = $fields['customFields'][$entity] ?? [];
                    $fields['customFields'][$entity][$customFieldName] = $value;
                } else {
                    $fields[$role] = $value;
                }
            } elseif ($field['type'] === 'checkbox') {
                $fields[$role] = false;
            }
        }

        $fields = $this->remapRebillyFields($fields);

        if (isset($fields['delivery_as_billing']) && $fields['delivery_as_billing']) {
            $fields['deliveryAddress'] = $fields['billingAddress'];
        }

        foreach ($fields as $key => $value) {
            if (is_array($value) && empty($value)) {
                unset($fields[$key]);
            }
        }

        return $fields;
    }

    private function remapRebillyFields(array $fields): array
    {
        foreach (self::FIELD_MAP as $sourceKey => $map) {
            if (!isset($fields[$sourceKey])) {
                continue;
            }
            $map = (array) $map;
            foreach ($map as $targetKey) {
                $path = explode('.', $targetKey);
                $target = &$fields;
                while (count($path) > 1) {
                    $target = &$target[array_shift($path)];
                }
                $target[$path[0]] = $fields[$sourceKey];
            }
            unset($fields[$sourceKey]);
        }

        return $fields;
    }

    public function getFieldAliases(): array
    {
        $aliases = [];
        foreach ($this->getFields() as $fieldName => $field) {
            $aliases[$fieldName] = $field['label'] ??
                mb_convert_case(str_replace(['_', '-', '.'], ' ', $fieldName), MB_CASE_TITLE);
        }

        return $aliases;
    }

    /**
     * @param mixed $value
     * @param array $customField
     * @return mixed
     */
    private function castCustomField($value, array $customField)
    {
        switch ($customField['type']) {
            case CustomField::TYPE_BOOLEAN:
                return in_array(strtolower(trim($value)), ['yes', 'y', 'true', '1', '+'], true);
            case CustomField::TYPE_INTEGER:
                return is_numeric($value) ? (int) $value : 0;
            case CustomField::TYPE_NUMBER:
                return is_numeric($value) ? (float) $value : 0;
        }

        return $value;
    }
}
