<?php
/**
 * This source file is proprietary and part of Rebilly.
 *
 * (c) Rebilly SRL
 *     Rebilly Ltd.
 *     Rebilly Inc.
 *
 * @see https://www.rebilly.com
 */

namespace Rebilly\Wp\Member;

use Rebilly\Wp\Member\Action\Action;
use Rebilly\Wp\Member\Admin\Admin;
use Rebilly\Wp\Member\Api\Client as Api;
use Rebilly\Wp\Member\Entity\Form;
use Rebilly\Wp\Member\Entity\RestrictableType;
use Rebilly\Wp\Member\Entity\Subscription;
use Rebilly\Wp\Member\Integration\Integration;
use Rebilly\Wp\Member\Shortcode\Shortcode;

// Exit if accessed directly.
if (!defined('ABSPATH')) {
    exit();
}

class Plugin
{
    /** @var Plugin|bool */
    private static $_instance = false;

    public $shortcode;

    /** @var Webhook */
    public $webhook;

    /** @var RulesEngine */
    public $rulesEngine;

    /** @var Admin */
    public $admin;

    private $api;

    private $loaded = false;

    private function __construct()
    {
        Helper::load();

        Settings::set('', apply_filters('rebilly_settings', get_option('wp_rebilly_m_settings')));
        do_action('rebilly_settings_loaded');

        do_action('rebilly_before_api');
        $this->api = new Api(
            Settings::get('secret_key', ''),
            Settings::get('website_id', ''),
            Settings::get('is_sandbox', false)
        );
        do_action('rebilly_after_api');

        if ($this->isConnected()) {
            Form::register();
            Subscription::register();
            Shortcode::register();
            Action::register();
            $this->webhook = new Webhook($this);
            $this->rulesEngine = new RulesEngine($this);

            Integration::register();
        }
    }

    public function isApiKeysValid(): bool
    {
        $isSandbox = Settings::get('is_sandbox', false);

        $secretKey = Settings::get('secret_key', '');
        $publishableKey = Settings::get('publishable_key', '');

        $secretKeyParts = explode('_', $secretKey, 3);
        $publishableKeyParts = explode('_', $publishableKey, 3);

        if (blank($secretKey) ||
            blank($publishableKey) ||
            count($secretKeyParts) !== 3 ||
            count($publishableKeyParts) !== 3) {
            return false;
        }

        [$secretType, $secretEnv,] = $secretKeyParts;
        [$publishableType, $publishableEnv,] = $publishableKeyParts;

        return $secretType === 'sk'
            && $publishableType === 'pk'
            && $secretEnv === $publishableEnv
            && $secretEnv === ($isSandbox ? 'sandbox' : 'live');
    }

    public function isConnected(): bool
    {
        return $this->isApiKeysValid() && $this->api->isConnected();
    }

    public function assets(): void
    {
        wp_register_style('rebilly-framepay', 'https://cdn.rebilly.com/framepay/v1/rebilly.css', false, '1.0.0');
        wp_register_script('rebilly-framepay', 'https://cdn.rebilly.com/framepay/v1/rebilly.js', ['jquery'], '1.0.0', true);

        wp_register_style('rebilly-front', REBILLY_WP_URL . 'dist/css/front.css', ['rebilly-framepay'], '1.0.0');
        wp_register_script('rebilly-front', REBILLY_WP_URL . 'dist/js/front.js', ['jquery', 'rebilly-framepay'], '1.0.0', true);

        wp_localize_script('rebilly-front', 'wp_rebilly_m', [
            'publishable_key' => Settings::get('publishable_key', ''),
            'ajax_url' => admin_url('admin-ajax.php'),
        ]);

        if (!is_admin()) {
            wp_enqueue_style('rebilly-framepay');
            wp_enqueue_script('rebilly-framepay');
            wp_enqueue_style('rebilly-front');
            wp_enqueue_script('rebilly-front');
        }
    }

    /**
     * @return static
     */
    public static function instance(): self
    {
        if (static::$_instance === false) {
            static::$_instance = new static();
        }
        static::$_instance->load();

        return static::$_instance;
    }

    public static function api(): Api
    {
        return static::instance()->api;
    }

    public function getApi(): Api
    {
        return $this->api;
    }

    public static function rulesEngine(): RulesEngine
    {
        return static::instance()->rulesEngine;
    }

    public function getRulesEngine(): RulesEngine
    {
        return $this->rulesEngine;
    }

    public function redirectNonAdmins(): void
    {
        if (!(defined('DOING_AJAX') && DOING_AJAX) &&
            get_current_user_id() !== 0 &&
            is_admin() &&
            (Settings::get('restrict_non_admins') ?? false) &&
            !current_user_can('administrator')) {
            $to = Settings::get('login_redirect');
            wp_redirect($to ? get_the_permalink($to) : home_url());
            exit;
        }
    }

    public function loginRedirect($redirect_to)
    {
        if ($to = Settings::get('login_redirect')) {
            return get_the_permalink($to);
        }

        return $redirect_to;
    }

    private function load(): void
    {
        if (!$this->loaded) {
            $this->loaded = true;
            $this->load_hooks();
            $this->load_admin();
        }
    }

    private function load_hooks(): void
    {
        do_action('rebilly_hook_load');

        add_action('wp_enqueue_scripts', [$this, 'assets']);
        add_filter('the_content', [$this, 'restrictPost']);
        add_filter('comments_open', [$this, 'restrictComments'], 10, 2);

        add_filter('login_redirect', [$this, 'loginRedirect']);
        add_action('init', [$this, 'redirectNonAdmins']);

        // TODO Filter query_vars
    }

    private function load_admin(): void
    {
        $this->admin = new Admin();
    }

    public function appUrl(): string
    {
        return Settings::get('is_sandbox', false)
            ? 'https://app-sandbox.rebilly.com/'
            : 'https://app.rebilly.com/';
    }

    public function isLocal(): bool
    {
        $localAddresses = [
            '127.0.0.1',
            '::1',
        ];

        return in_array($this->getIp(), $localAddresses, true);
    }

    public function restrictPost($content = null)
    {
        $post = RestrictableType::find();
        if ($post && !$post->canAccess()) {
            $content = Settings::get('message_restricted');
        }

        return $content;
    }

    public function restrictComments($open, $postId): bool
    {
        $post = RestrictableType::find($postId);
        if ($post && !$post->canAccess()) {
            $open = false;
        }

        return $open;
    }

    public function uninstall(): void
    {
        if ($this->isConnected()) {
            $this->webhook->uninstall();
        }
    }

    private function getIp(): string
    {
        if (isset($_SERVER['HTTP_CLIENT_IP'])) {
            $ip = $_SERVER['HTTP_CLIENT_IP'];
        } elseif (isset($_SERVER['HTTP_X_FORWARDED_FOR'])) {
            $ip = $_SERVER['HTTP_X_FORWARDED_FOR'];
        } elseif (isset($_SERVER['HTTP_X_FORWARDED'])) {
            $ip = $_SERVER['HTTP_X_FORWARDED'];
        } elseif (isset($_SERVER['HTTP_FORWARDED_FOR'])) {
            $ip = $_SERVER['HTTP_FORWARDED_FOR'];
        } elseif (isset($_SERVER['HTTP_FORWARDED'])) {
            $ip = $_SERVER['HTTP_FORWARDED'];
        } else {
            $ip = $_SERVER['REMOTE_ADDR'] ?? '127.0.0.1';
        }

        return $ip;
    }
}
