<?php
/**
 * This source file is proprietary and part of Rebilly.
 *
 * (c) Rebilly SRL
 *     Rebilly Ltd.
 *     Rebilly Inc.
 *
 * @see https://www.rebilly.com
 */

namespace Rebilly\Wp\Member;

use Exception;
use Rebilly\Http\Exception\UnprocessableEntityException;
use Rebilly\Wp\Member\Entity\User;

// Exit if accessed directly.
if (!defined('ABSPATH')) {
    exit();
}

class Webhook
{
    private $settings;

    private $settings_key = 'wp_rebilly_m_webhook_settings';

    private $plugin;

    public function __construct(Plugin $plugin)
    {
        $this->plugin = $plugin;
        $this->settings = apply_filters('rebilly_webhook_settings', get_option($this->settings_key, []));
        if (!is_array($this->settings)) {
            $this->settings = [];
            update_option('wp_rebilly_m_webhook_settings', $this->settings);
        }

        if ($plugin->isConnected()) {
            $this->loadHooks();
        }
    }

    public function loadHooks(): void
    {
        if ($this->plugin->isLocal()) {
            rebilly_admin_notice(__('Rebilly Webhook is not available for local environment.', REBILLY_WP_LANG_DOMAIN), 'warning', true);
        } else {
            if (!$this->isSetUp() && !$this->install()) {
                rebilly_admin_notice(__('Failed to add Rebilly Webhook. Please, check API keys and contact support.', REBILLY_WP_LANG_DOMAIN), 'error', true);

                return;
            }
            $action = $this->settings('action');
            add_action("wp_ajax_$action", [$this, 'listen']);
            add_action("wp_ajax_nopriv_$action", [$this, 'listen']);
        }
    }

    private function isSetUp(): bool
    {
        $required = ['webhookId', 'events', 'credentialHash', 'action', 'username', 'password'];
        foreach ($required as $key) {
            if (blank($this->settings($key))) {
                return false;
            }
        }
        $events = $this->settings('events');

        return $events === sha1(json_encode($this->requiredEvents()));
    }

    private function settings($key = null, $default = null)
    {
        if ($key === null) {
            return $this->settings;
        }
        if (is_array($key)) {
            $this->settings = $default === true ? $key : array_merge($this->settings, $key);

            return update_option($this->settings_key, $this->settings);
        }

        return $this->settings[$key] ?? $default;
    }

    public function install(): bool
    {
        if (!$this->plugin->isConnected() ||
            $this->plugin->isLocal() ||
            $this->isSetUp() ||
            get_transient('rebilly-webhook-delay') !== false
        ) {
            return false;
        }

        if (blank(
            $this->settings('action'),
            $this->settings('username'),
            $this->settings('password'),
            $this->settings('credentialHash')
        )) {
            $settings = [
                'action' => uniqid('rebilly_hook_', false),
                'username' => uniqid('', true),
                'password' => uniqid('', true),
            ];

            try {
                $settings['credentialHash'] = $this->createCredentials($settings['username'], $settings['password']);
            } catch (Exception $ex) {
                set_transient('rebilly-webhook-delay', true, 5 * 60);

                return false;
            }
            $this->settings($settings);
        }
        if (
            blank($this->settings('webhookId'), $this->settings('events')) ||
            $this->settings('events') !== sha1(json_encode($this->requiredEvents()))
        ) {
            try {
                $settings = [
                    'webhookId' => $this->createWebhook($this->settings('action'), $this->settings('credentialHash')),
                    'events' => sha1(json_encode($this->requiredEvents())),
                ];
            } catch (Exception $ex) {
                set_transient('rebilly-webhook-delay', true, 5 * 60);

                return false;
            }
            $this->settings($settings);
        }

        return true;
    }

    /**
     * @param $username
     * @param $password
     *
     * @throws UnprocessableEntityException
     * @return bool|string
     */
    private function createCredentials($username, $password)
    {
        if (!$this->plugin->isConnected()) {
            return false;
        }
        $credential = $this->plugin->getApi()
            ->webhook()
            ->createCredential([
                'host' => parse_url(site_url(), PHP_URL_HOST),
                'auth' => [
                    'type' => 'basic',
                    'username' => $username,
                    'password' => $password,
                ],
            ]);

        /** @noinspection NullPointerExceptionInspection */
        return $credential->getId();
    }

    /**
     * @param $action
     * @param $credentialHash
     *
     * @throws UnprocessableEntityException
     * @return bool|string
     */
    private function createWebhook($action, $credentialHash)
    {
        if (!$this->plugin->isConnected()) {
            return false;
        }

        $websiteId = Settings::get('website_id');
        $webhookId = $this->settings('webhookId', "wp-{$websiteId}");
        $webhookId = preg_replace('/[^A-Za-z0-9_-]/', '_', $webhookId);

        $webhook = $this->plugin->getApi()
            ->webhook()
            ->update($webhookId, [
                'eventsFilter' => $this->requiredEvents(),
                'method' => 'POST',
                'url' => admin_url("admin-ajax.php?action={$action}"),
                'credentialHash' => $credentialHash,
            ]);

        /** @noinspection NullPointerExceptionInspection */
        return $webhook->getId();
    }

    private function requiredEvents(): array
    {
        $events = ['subscription-activated', 'subscription-canceled', 'subscription-renewed'];
        if (Settings::get('auto_register', false)) {
            $events[] = 'customer-created';
        }

        return $events;
    }

    public function uninstall(): void
    {
        if ($this->plugin->isConnected() && $this->isSetUp()) {
            try {
                $webhook = $this->plugin->getApi()
                    ->webhook()
                    ->get($this->settings('webhookId'));
                if ($webhook) {
                    $webhook->setStatus('inactive');

                    $this->plugin->getApi()
                        ->webhook()
                        ->update($webhook->getId(), $webhook);
                }

            } catch (UnprocessableEntityException $e) {
                // Nothing we can do
            }
        }
        delete_option($this->settings_key);
    }

    public function listen(): void
    {
        $this->auth();

        $input = @file_get_contents('php://input');
        $data = json_decode((string) $input, true);

        $handlers = [
            'customer-created' => [$this, 'handleCustomerCreated'],
            'subscription-activated' => [$this, 'handleSubscriptionActivated'],
            'subscription-renewed' => [$this, 'handleSubscriptionRenewed'],
            'subscription-canceled' => [$this, 'handleSubscriptionCanceled'],
        ];
        $event = $data['eventType'];
        if (array_key_exists($event, $handlers) && is_callable($handlers[$event])) {
            call_user_func($handlers[$event], $data);
        }

        wp_die();
    }

    private function auth(): void
    {
        if (!(isset($_SERVER['PHP_AUTH_USER'], $_SERVER['PHP_AUTH_PW'])
            && $_SERVER['PHP_AUTH_USER'] === $this->settings('username')
            && $_SERVER['PHP_AUTH_PW'] === $this->settings('password'))) {
            header('WWW-Authenticate: Basic realm="Restricted area"');
            header('HTTP/1.1 401 Unauthorized');
            exit;
        }
    }

    public function handleCustomerCreated(array $data): bool
    {
        $customer = $data['_embedded']['customer'];

        $user = User::get_data_by('email', $customer['email']);
        if ($user !== false) {
            $user = User::find($user);
            if (blank($user->getCustomerId())) {
                $user->setCustomerId($customer['id']);
            }
        } else {
            $user = User::create([
                'email' => $customer['email'],
                'first_name' => $customer['firstName'],
                'last_name' => $customer['lastName'],
            ]);
            if (is_wp_error($user)) {
                return false;
            }
            $user->setCustomerId($customer['id']);

            if (Settings::get('auto_register_send_confirmation', false)) {
                wp_new_user_notification($user, null, 'user');
            }
        }

        return true;
    }

    public function handleSubscriptionRenewed(array $data): bool
    {
        return $this->handleSubscriptionActivated($data);
    }

    public function handleSubscriptionActivated(array $data): bool
    {
        $subscription = $data['_embedded']['subscription'];
        if (!($user = User::findByCustomerId($subscription['customerId']))) {
            return false;
        }
        $capabilities = [];
        foreach ($subscription['items'] as $item) {
            $capabilities[] = $item['id'];
        }

        $user->grantCapability($capabilities);

        return true;
    }

    public function handleSubscriptionCanceled(array $data): bool
    {
        $subscription = $data['_embedded']['subscription'];
        if (!($user = User::findByCustomerId($subscription['customerId']))) {
            return false;
        }
        $capabilities = [];
        foreach ($subscription['items'] as $item) {
            $capabilities[] = $item['id'];
        }

        $user->revokeCapability($capabilities);

        return true;
    }
}
